'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

exports.default = buildProgram;

var _babelTypes = require('babel-types');

var t = _interopRequireWildcard(_babelTypes);

var _babelPluginFlowRuntime = require('babel-plugin-flow-runtime');

var _crawlTypeDependencies = require('./crawlTypeDependencies');

var _Graph = require('./Graph');

var _flowConfigParser = require('flow-config-parser');

var _flowRuntime = require('flow-runtime');

var _flowRuntime2 = _interopRequireDefault(_flowRuntime);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var FlowConfig = _flowRuntime2.default.tdz(function () {
  return _flowConfigParser.FlowConfig;
});

var FlowModule = _flowRuntime2.default.tdz(function () {
  return _Graph.FlowModule;
});

var DependencyDict = _flowRuntime2.default.tdz(function () {
  return _crawlTypeDependencies.DependencyDict;
});

var ModuleDependencyDict = _flowRuntime2.default.tdz(function () {
  return _crawlTypeDependencies.ModuleDependencyDict;
});

var CrawledDependencies = _flowRuntime2.default.tdz(function () {
  return _crawlTypeDependencies.CrawledDependencies;
});

var Node = _flowRuntime2.default.type('Node', _flowRuntime2.default.object(_flowRuntime2.default.property('type', _flowRuntime2.default.string())));

function buildProgram(config, _arg, graph) {
  var _configType = _flowRuntime2.default.nullable(_flowRuntime2.default.ref(FlowConfig));

  var _graphType = _flowRuntime2.default.ref(FlowModule);

  var _returnType = _flowRuntime2.default.return(Node);

  _flowRuntime2.default.warn(_flowRuntime2.default.param('config', _configType), config);

  _flowRuntime2.default.warn(_flowRuntime2.default.param('graph', _graphType), graph);

  var _t$warn = _flowRuntime2.default.warn(_flowRuntime2.default.ref(CrawledDependencies), _arg),
      globalDependencies = _t$warn.globalDependencies,
      moduleDependencies = _t$warn.moduleDependencies;

  var modules = _flowRuntime2.default.warn(_flowRuntime2.default.ref('Map', _flowRuntime2.default.ref(FlowModule), _flowRuntime2.default.array(_flowRuntime2.default.object())), new Map());
  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = graph.traverse.apply(graph, _toConsumableArray(toKeys(globalDependencies, moduleDependencies)))[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var vertex = _step.value;

      if (vertex.path && vertex.path.type !== 'TypeParameter') {
        var node = vertex.path.node;

        var existing = modules.get(vertex.parent);
        if (existing) {
          existing.push(node);
        } else {
          modules.set(vertex.parent, [node]);
        }
      }
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  var prelude = makePrelude(config);
  var block = [].concat(_toConsumableArray(prelude));
  var _iteratorNormalCompletion2 = true;
  var _didIteratorError2 = false;
  var _iteratorError2 = undefined;

  try {
    for (var _iterator2 = modules[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
      var _ref = _step2.value;

      var _ref2 = _slicedToArray(_ref, 2);

      var mod = _ref2[0];
      var items = _ref2[1];

      if (mod.name) {
        block.push(t.declareModule(t.stringLiteral(mod.name), t.blockStatement(items)));
      } else {
        block.push.apply(block, _toConsumableArray(items));
      }
    }
  } catch (err) {
    _didIteratorError2 = true;
    _iteratorError2 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion2 && _iterator2.return) {
        _iterator2.return();
      }
    } finally {
      if (_didIteratorError2) {
        throw _iteratorError2;
      }
    }
  }

  var prog = t.file(t.program(block));
  (0, _babelPluginFlowRuntime.transform)(prog);
  return _flowRuntime2.default.warn(_returnType, prog);
}

_flowRuntime2.default.annotate(buildProgram, _flowRuntime2.default.function(_flowRuntime2.default.param('config', _flowRuntime2.default.nullable(_flowRuntime2.default.ref(FlowConfig))), _flowRuntime2.default.param('_arg1', _flowRuntime2.default.ref(CrawledDependencies)), _flowRuntime2.default.param('graph', _flowRuntime2.default.ref(FlowModule)), _flowRuntime2.default.return(Node)));

function makePrelude(config) {
  var _configType2 = _flowRuntime2.default.nullable(_flowRuntime2.default.ref(FlowConfig));

  _flowRuntime2.default.warn(_flowRuntime2.default.param('config', _configType2), config);

  var prelude = [];
  if (!config) {
    return prelude;
  }
  var suppressionTypeNames = config.get('suppress_type');
  if (suppressionTypeNames.length > 0) {
    var aliases = [];
    var _iteratorNormalCompletion3 = true;
    var _didIteratorError3 = false;
    var _iteratorError3 = undefined;

    try {
      for (var _iterator3 = suppressionTypeNames[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
        var typeName = _step3.value;

        if (typeName === '$FlowFixMe') {
          continue; // skip the builtin
        }
        aliases.push(t.expressionStatement(t.callExpression(t.memberExpression(t.identifier('t'), t.identifier('declare')), [t.callExpression(t.memberExpression(t.identifier('t'), t.identifier('type')), [t.stringLiteral(typeName), t.callExpression(t.memberExpression(t.identifier('t'), t.identifier('$flowFixMe')), [])])])));
      }
    } catch (err) {
      _didIteratorError3 = true;
      _iteratorError3 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion3 && _iterator3.return) {
          _iterator3.return();
        }
      } finally {
        if (_didIteratorError3) {
          throw _iteratorError3;
        }
      }
    }

    if (aliases.length > 0) {
      prelude.push(t.importDeclaration([t.importDefaultSpecifier(t.identifier('t'))], t.stringLiteral('flow-runtime')));
      prelude.push.apply(prelude, aliases);
    }
  }
  return prelude;
}

_flowRuntime2.default.annotate(makePrelude, _flowRuntime2.default.function(_flowRuntime2.default.param('config', _flowRuntime2.default.nullable(_flowRuntime2.default.ref(FlowConfig)))));

function toKeys(globalDependencies, moduleDependencies) {
  var _globalDependenciesType = _flowRuntime2.default.ref(DependencyDict);

  var _moduleDependenciesType = _flowRuntime2.default.ref(ModuleDependencyDict);

  var _returnType2 = _flowRuntime2.default.return(_flowRuntime2.default.array(_flowRuntime2.default.union(_flowRuntime2.default.string(), _flowRuntime2.default.array(_flowRuntime2.default.string()))));

  _flowRuntime2.default.warn(_flowRuntime2.default.param('globalDependencies', _globalDependenciesType), globalDependencies);

  _flowRuntime2.default.warn(_flowRuntime2.default.param('moduleDependencies', _moduleDependenciesType), moduleDependencies);

  var keys = Object.keys(globalDependencies);
  for (var moduleName in moduleDependencies) {
    var module = moduleDependencies[moduleName];
    for (var key in module) {
      keys.push([moduleName, key]);
    }
  }
  return _flowRuntime2.default.warn(_returnType2, keys);
}

_flowRuntime2.default.annotate(toKeys, _flowRuntime2.default.function(_flowRuntime2.default.param('globalDependencies', _flowRuntime2.default.ref(DependencyDict)), _flowRuntime2.default.param('moduleDependencies', _flowRuntime2.default.ref(ModuleDependencyDict)), _flowRuntime2.default.return(_flowRuntime2.default.array(_flowRuntime2.default.union(_flowRuntime2.default.string(), _flowRuntime2.default.array(_flowRuntime2.default.string()))))));