'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = collectExternalTypeReferences;

var _babelTraverse = require('babel-traverse');

var _babelTraverse2 = _interopRequireDefault(_babelTraverse);

var _babelPluginFlowRuntime = require('babel-plugin-flow-runtime');

var _shouldIgnoreType = require('./shouldIgnoreType');

var _shouldIgnoreType2 = _interopRequireDefault(_shouldIgnoreType);

var _flowRuntime = require('flow-runtime');

var _flowRuntime2 = _interopRequireDefault(_flowRuntime);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var NodePath = _flowRuntime2.default.tdz(function () {
  return _babelTraverse.NodePath;
});

var Scope = _flowRuntime2.default.tdz(function () {
  return _babelTraverse.Scope;
});

var Node = _flowRuntime2.default.type('Node', _flowRuntime2.default.object(_flowRuntime2.default.property('type', _flowRuntime2.default.string())));

var nodeTypeParameters = new WeakMap();

function collectExternalTypeReferences(file) {
  var _fileType = Node;

  _flowRuntime2.default.warn(_flowRuntime2.default.param('file', _fileType), file);

  var importedTypes = {};
  var globalTypes = {};

  // First pass, collect value and type definitions
  (0, _babelTraverse2.default)(file, {
    TypeAlias(path) {
      var _pathType = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType), path);

      registerDefinition(path.get('id'), 'type');
      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = (0, _babelPluginFlowRuntime.getTypeParameters)(path)[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          var typeParameter = _step.value;

          registerTypeParameter(typeParameter, path);
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }
    },
    InterfaceDeclaration(path) {
      var _pathType2 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType2), path);

      registerDefinition(path.get('id'), 'type');
      var _iteratorNormalCompletion2 = true;
      var _didIteratorError2 = false;
      var _iteratorError2 = undefined;

      try {
        for (var _iterator2 = (0, _babelPluginFlowRuntime.getTypeParameters)(path)[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
          var typeParameter = _step2.value;

          registerTypeParameter(typeParameter, path);
        }
      } catch (err) {
        _didIteratorError2 = true;
        _iteratorError2 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion2 && _iterator2.return) {
            _iterator2.return();
          }
        } finally {
          if (_didIteratorError2) {
            throw _iteratorError2;
          }
        }
      }
    },
    ImportDeclaration(path) {
      var _pathType3 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType3), path);

      path.get('specifiers').forEach(function (specifier) {
        var local = specifier.get('local');
        if (path.node.importKind === 'type') {
          registerDefinition(local, 'importedType');
        } else {
          registerDefinition(local, 'value');
        }
      });
    },
    VariableDeclarator(path) {
      var _pathType4 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType4), path);

      var _iteratorNormalCompletion3 = true;
      var _didIteratorError3 = false;
      var _iteratorError3 = undefined;

      try {
        for (var _iterator3 = (0, _babelPluginFlowRuntime.findIdentifiers)(path.get('id'))[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
          var id = _step3.value;

          registerDefinition(id, 'value');
        }
      } catch (err) {
        _didIteratorError3 = true;
        _iteratorError3 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion3 && _iterator3.return) {
            _iterator3.return();
          }
        } finally {
          if (_didIteratorError3) {
            throw _iteratorError3;
          }
        }
      }
    },
    Function(path) {
      var _pathType5 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType5), path);

      if (path.isFunctionDeclaration() && path.has('id')) {
        registerDefinition(path.get('id'), 'value', path.parentPath.scope);
      }

      var _iteratorNormalCompletion4 = true;
      var _didIteratorError4 = false;
      var _iteratorError4 = undefined;

      try {
        for (var _iterator4 = (0, _babelPluginFlowRuntime.findIdentifiers)(path.get('params'))[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
          var id = _step4.value;

          registerDefinition(id, 'value');
        }
      } catch (err) {
        _didIteratorError4 = true;
        _iteratorError4 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion4 && _iterator4.return) {
            _iterator4.return();
          }
        } finally {
          if (_didIteratorError4) {
            throw _iteratorError4;
          }
        }
      }

      var _iteratorNormalCompletion5 = true;
      var _didIteratorError5 = false;
      var _iteratorError5 = undefined;

      try {
        for (var _iterator5 = (0, _babelPluginFlowRuntime.getTypeParameters)(path)[Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
          var typeParameter = _step5.value;

          registerTypeParameter(typeParameter, path);
        }
      } catch (err) {
        _didIteratorError5 = true;
        _iteratorError5 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion5 && _iterator5.return) {
            _iterator5.return();
          }
        } finally {
          if (_didIteratorError5) {
            throw _iteratorError5;
          }
        }
      }
    },
    Class(path) {
      var _pathType6 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType6), path);

      if (path.isClassDeclaration() && path.has('id')) {
        registerDefinition(path.get('id'), 'value', path.parentPath.scope);
      }

      var _iteratorNormalCompletion6 = true;
      var _didIteratorError6 = false;
      var _iteratorError6 = undefined;

      try {
        for (var _iterator6 = (0, _babelPluginFlowRuntime.getTypeParameters)(path)[Symbol.iterator](), _step6; !(_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done); _iteratorNormalCompletion6 = true) {
          var typeParameter = _step6.value;

          registerTypeParameter(typeParameter, path);
        }
      } catch (err) {
        _didIteratorError6 = true;
        _iteratorError6 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion6 && _iterator6.return) {
            _iterator6.return();
          }
        } finally {
          if (_didIteratorError6) {
            throw _iteratorError6;
          }
        }
      }
    }
  });

  (0, _babelTraverse2.default)(file, {
    ImportDeclaration(path) {
      var _pathType7 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType7), path);

      path.get('specifiers').forEach(function (specifier) {
        var source = path.get('source').node.value;
        var local = specifier.get('local');
        var name = local.node.name;

        if (path.node.importKind === 'type') {
          var isDefault = specifier.isImportDefaultSpecifier();
          var isNamespace = !isDefault && specifier.isImportNamespaceSpecifier();
          var imported = isDefault || isNamespace ? name : specifier.get('imported').node.name;
          importedTypes[name] = {
            imported,
            source,
            isDefault,
            isNamespace
          };
        }
      });
    },
    GenericTypeAnnotation(path) {
      var _pathType8 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType8), path);

      var id = path.get('id');
      var name = id.node.name;

      var kind = getDefinitionKind(id);
      if (!kind && !(0, _shouldIgnoreType2.default)(name)) {
        if (globalTypes[name]) {
          globalTypes[name]++;
        } else {
          globalTypes[name] = 1;
        }
      }
    }
  });

  return { importedTypes, globalTypes };
}

_flowRuntime2.default.annotate(collectExternalTypeReferences, _flowRuntime2.default.function(_flowRuntime2.default.param('file', Node)));

function getDefinitionKind(id) {
  var _idType = _flowRuntime2.default.ref(NodePath);

  var _returnType = _flowRuntime2.default.return(_flowRuntime2.default.union(_flowRuntime2.default.nullable(_flowRuntime2.default.string('value')), _flowRuntime2.default.string('type'), _flowRuntime2.default.string('importedType'), _flowRuntime2.default.string('typeParameter')));

  _flowRuntime2.default.warn(_flowRuntime2.default.param('id', _idType), id);

  var fromScope = id.scope.getData(`entityKind:${id.node.name}`);
  if (fromScope) {
    return _flowRuntime2.default.warn(_returnType, fromScope);
  }
  var parent = id.parentPath;
  while (parent && parent.node) {
    var typeParams = nodeTypeParameters.get(parent.node);
    if (typeParams && typeParams[id.node.name]) {
      return _flowRuntime2.default.warn(_returnType, 'typeParameter');
    }
    parent = parent.parentPath;
  }
}

_flowRuntime2.default.annotate(getDefinitionKind, _flowRuntime2.default.function(_flowRuntime2.default.param('id', _flowRuntime2.default.ref(NodePath)), _flowRuntime2.default.return(_flowRuntime2.default.union(_flowRuntime2.default.nullable(_flowRuntime2.default.string('value')), _flowRuntime2.default.string('type'), _flowRuntime2.default.string('importedType'), _flowRuntime2.default.string('typeParameter')))));

function registerTypeParameter(id, owner) {
  var _idType2 = _flowRuntime2.default.ref(NodePath);

  var _ownerType = _flowRuntime2.default.ref(NodePath);

  _flowRuntime2.default.warn(_flowRuntime2.default.param('id', _idType2), id);

  _flowRuntime2.default.warn(_flowRuntime2.default.param('owner', _ownerType), owner);

  var items = nodeTypeParameters.get(owner.node);
  if (!items) {
    items = {};
    nodeTypeParameters.set(owner.node, items);
  }
  items[id.node.name] = true;
}

_flowRuntime2.default.annotate(registerTypeParameter, _flowRuntime2.default.function(_flowRuntime2.default.param('id', _flowRuntime2.default.ref(NodePath)), _flowRuntime2.default.param('owner', _flowRuntime2.default.ref(NodePath))));

function registerDefinition(id, kind) {
  var scope = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : id.scope;

  var _idType3 = _flowRuntime2.default.ref(NodePath);

  var _kindType = _flowRuntime2.default.union(_flowRuntime2.default.string('value'), _flowRuntime2.default.string('type'), _flowRuntime2.default.string('importedType'));

  var _scopeType = _flowRuntime2.default.ref(Scope);

  _flowRuntime2.default.warn(_flowRuntime2.default.param('id', _idType3), id);

  _flowRuntime2.default.warn(_flowRuntime2.default.param('kind', _kindType), kind);

  _flowRuntime2.default.warn(_flowRuntime2.default.param('scope', _scopeType), scope);

  scope.setData(`entityKind:${id.node.name}`, kind);
}

_flowRuntime2.default.annotate(registerDefinition, _flowRuntime2.default.function(_flowRuntime2.default.param('id', _flowRuntime2.default.ref(NodePath)), _flowRuntime2.default.param('kind', _flowRuntime2.default.union(_flowRuntime2.default.string('value'), _flowRuntime2.default.string('type'), _flowRuntime2.default.string('importedType'))), _flowRuntime2.default.param('scope', _flowRuntime2.default.ref(Scope))));