'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = fastTraverse;

var _babelTypes = require('babel-types');

var t = _interopRequireWildcard(_babelTypes);

var _doubleEndedQueue = require('double-ended-queue');

var _doubleEndedQueue2 = _interopRequireDefault(_doubleEndedQueue);

var _flowRuntime = require('flow-runtime');

var _flowRuntime2 = _interopRequireDefault(_flowRuntime);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

var Node = _flowRuntime2.default.type('Node', _flowRuntime2.default.object(_flowRuntime2.default.property('type', _flowRuntime2.default.string()), _flowRuntime2.default.indexer('key', _flowRuntime2.default.union(_flowRuntime2.default.number(), _flowRuntime2.default.string()), _flowRuntime2.default.any())));

var NodeVisitor = _flowRuntime2.default.type('NodeVisitor', _flowRuntime2.default.function(_flowRuntime2.default.param('node', Node), _flowRuntime2.default.param('key', _flowRuntime2.default.union(_flowRuntime2.default.nullable(_flowRuntime2.default.number()), _flowRuntime2.default.string())), _flowRuntime2.default.param('listKey', _flowRuntime2.default.nullable(_flowRuntime2.default.string())), _flowRuntime2.default.param('parent', _flowRuntime2.default.nullable(Node)), _flowRuntime2.default.return(_flowRuntime2.default.any())));

var VISITOR_KEYS = Object.assign({}, t.VISITOR_KEYS, {
  NumberLiteralTypeAnnotation: [],
  ExistsTypeAnnotation: [],
  DeclareExportDeclaration: ['declaration']
});

function fastTraverse(ast, visitor) {
  var _astType = Node;
  var _visitorType = NodeVisitor;

  _flowRuntime2.default.warn(_flowRuntime2.default.param('ast', _astType), ast);

  _flowRuntime2.default.warn(_flowRuntime2.default.param('visitor', _visitorType), visitor);

  traverseNode(ast, null, null, null, visitor);
}

_flowRuntime2.default.annotate(fastTraverse, _flowRuntime2.default.function(_flowRuntime2.default.param('ast', Node), _flowRuntime2.default.param('visitor', NodeVisitor)));

function traverseNode(node, key, listKey, parent, visitor) {
  var _nodeType = Node;

  var _keyType = _flowRuntime2.default.union(_flowRuntime2.default.nullable(_flowRuntime2.default.number()), _flowRuntime2.default.string());

  var _listKeyType = _flowRuntime2.default.nullable(_flowRuntime2.default.string());

  var _parentType = _flowRuntime2.default.nullable(Node);

  var _visitorType2 = NodeVisitor;

  _flowRuntime2.default.warn(_flowRuntime2.default.param('node', _nodeType), node);

  _flowRuntime2.default.warn(_flowRuntime2.default.param('key', _keyType), key);

  _flowRuntime2.default.warn(_flowRuntime2.default.param('listKey', _listKeyType), listKey);

  _flowRuntime2.default.warn(_flowRuntime2.default.param('parent', _parentType), parent);

  _flowRuntime2.default.warn(_flowRuntime2.default.param('visitor', _visitorType2), visitor);

  var result = visitor(node, key, listKey, parent);
  if (result === false) {
    return false;
  }
  var keys = VISITOR_KEYS[node.type];
  if (keys && keys.length > 0) {
    for (var i = 0; i < keys.length; i++) {
      var _key = keys[i];
      var value = node[_key];
      if (Array.isArray(value)) {
        for (var j = 0; j < value.length; j++) {
          var item = value[j];
          if (item) {
            var _result = traverseNode(item, j, _key, node, visitor);
            if (_result === false) {
              return false;
            }
          }
        }
      } else if (value) {
        var _result2 = traverseNode(value, _key, _key, node, visitor);
        if (_result2 === false) {
          return false;
        }
      }
    }
  }
}

_flowRuntime2.default.annotate(traverseNode, _flowRuntime2.default.function(_flowRuntime2.default.param('node', Node), _flowRuntime2.default.param('key', _flowRuntime2.default.union(_flowRuntime2.default.nullable(_flowRuntime2.default.number()), _flowRuntime2.default.string())), _flowRuntime2.default.param('listKey', _flowRuntime2.default.nullable(_flowRuntime2.default.string())), _flowRuntime2.default.param('parent', _flowRuntime2.default.nullable(Node)), _flowRuntime2.default.param('visitor', NodeVisitor)));