'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = importAST;

var _babelTraverse = require('babel-traverse');

var _babelTraverse2 = _interopRequireDefault(_babelTraverse);

var _babelTypes = require('babel-types');

var t = _interopRequireWildcard(_babelTypes);

var _babelPluginFlowRuntime = require('babel-plugin-flow-runtime');

var _shouldIgnoreType = require('./shouldIgnoreType');

var _shouldIgnoreType2 = _interopRequireDefault(_shouldIgnoreType);

var _Graph = require('./Graph');

var _flowRuntime = require('flow-runtime');

var _flowRuntime2 = _interopRequireDefault(_flowRuntime);

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var FlowModule = _flowRuntime2.default.tdz(function () {
  return _Graph.FlowModule;
});

var FlowEntity = _flowRuntime2.default.tdz(function () {
  return _Graph.FlowEntity;
});

var NodePath = _flowRuntime2.default.tdz(function () {
  return _babelTraverse.NodePath;
});

var Scope = _flowRuntime2.default.tdz(function () {
  return _babelTraverse.Scope;
});

var Node = _flowRuntime2.default.type('Node', _flowRuntime2.default.object(_flowRuntime2.default.property('type', _flowRuntime2.default.string())));

function importAST(graph, file) {
  var _graphType = _flowRuntime2.default.ref(FlowModule);

  var _fileType = Node;

  _flowRuntime2.default.warn(_flowRuntime2.default.param('graph', _graphType), graph);

  _flowRuntime2.default.warn(_flowRuntime2.default.param('file', _fileType), file);

  var nodeTypeParameters = new WeakMap();

  var moduleStack = [graph];

  function currentModule() {
    var _returnType = _flowRuntime2.default.return(_flowRuntime2.default.ref(FlowModule));

    return _flowRuntime2.default.warn(_returnType, moduleStack[moduleStack.length - 1]);
  }

  _flowRuntime2.default.annotate(currentModule, _flowRuntime2.default.function(_flowRuntime2.default.return(_flowRuntime2.default.ref(FlowModule))));

  function getDefinition(id) {
    var _idType = _flowRuntime2.default.ref(NodePath);

    var _returnType2 = _flowRuntime2.default.return(_flowRuntime2.default.nullable(_flowRuntime2.default.ref(FlowEntity)));

    _flowRuntime2.default.warn(_flowRuntime2.default.param('id', _idType), id);

    var name = id.node.name;

    if ((0, _shouldIgnoreType2.default)(name)) {
      return _flowRuntime2.default.warn(_returnType2);
    }
    var fromScope = id.scope.getData(`entityDefinition:${name}`);
    if (fromScope) {
      return _flowRuntime2.default.warn(_returnType2, fromScope);
    }
    var parent = id.parentPath;
    while (parent && parent.node) {
      var typeParameters = nodeTypeParameters.get(parent.node);
      if (typeParameters) {}
      if (typeParameters && typeParameters[id.node.name]) {
        return _flowRuntime2.default.warn(_returnType2, typeParameters[id.node.name]);
      }
      parent = parent.parentPath;
    }
  }

  _flowRuntime2.default.annotate(getDefinition, _flowRuntime2.default.function(_flowRuntime2.default.param('id', _flowRuntime2.default.ref(NodePath)), _flowRuntime2.default.return(_flowRuntime2.default.nullable(_flowRuntime2.default.ref(FlowEntity)))));

  function registerTypeParameter(id, owner) {
    var _idType2 = _flowRuntime2.default.ref(NodePath);

    var _ownerType = _flowRuntime2.default.ref(NodePath);

    var _returnType3 = _flowRuntime2.default.return(_flowRuntime2.default.ref(FlowEntity));

    _flowRuntime2.default.warn(_flowRuntime2.default.param('id', _idType2), id);

    _flowRuntime2.default.warn(_flowRuntime2.default.param('owner', _ownerType), owner);

    var items = nodeTypeParameters.get(owner.node);
    if (!items) {
      items = {};
      nodeTypeParameters.set(owner.node, items);
    }
    var vertex = currentModule().getEntityForPath(id);
    items[id.node.name] = vertex;
    return _flowRuntime2.default.warn(_returnType3, vertex);
  }

  _flowRuntime2.default.annotate(registerTypeParameter, _flowRuntime2.default.function(_flowRuntime2.default.param('id', _flowRuntime2.default.ref(NodePath)), _flowRuntime2.default.param('owner', _flowRuntime2.default.ref(NodePath)), _flowRuntime2.default.return(_flowRuntime2.default.ref(FlowEntity))));

  function registerDefinition(path) {
    var scope = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : path.scope;

    var _pathType = _flowRuntime2.default.ref(NodePath);

    var _scopeType = _flowRuntime2.default.ref(Scope);

    var _returnType4 = _flowRuntime2.default.return(_flowRuntime2.default.ref(FlowEntity));

    _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType), path);

    _flowRuntime2.default.warn(_flowRuntime2.default.param('scope', _scopeType), scope);

    var shouldDeclare = path.type === 'DeclareVariable' || path.type === 'DeclareClass' || path.type === 'DeclareFunction' || path.type === 'DeclareModuleExports' || path.type === 'DeclareTypeAlias' || path.parentPath.type === 'ImportDeclaration' || path.parentPath.type === 'ImportDefaultDeclaration' || path.parentPath.type === 'ImportNamespaceDeclaration';
    var current = currentModule();
    var vertex = shouldDeclare ? current.register(path) : current.getEntityForPath(path);
    if (!path.isDeclareModuleExports()) {
      var id = path.isIdentifier() ? path : path.get('id');
      scope.setData(`entityDefinition:${id.node.name}`, vertex);
    }
    var typeParameters = (0, _babelPluginFlowRuntime.getTypeParameters)(path);
    if (typeParameters.length > 0) {
      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = typeParameters[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          var typeParameter = _step.value;

          registerTypeParameter(typeParameter, path);
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }
    }
    return _flowRuntime2.default.warn(_returnType4, vertex);
  }

  _flowRuntime2.default.annotate(registerDefinition, _flowRuntime2.default.function(_flowRuntime2.default.param('path', _flowRuntime2.default.ref(NodePath)), _flowRuntime2.default.param('scope', _flowRuntime2.default.ref(Scope)), _flowRuntime2.default.return(_flowRuntime2.default.ref(FlowEntity))));

  function resolvePossiblyQualified(path) {
    var _pathType2 = _flowRuntime2.default.ref(NodePath);

    var _returnType5 = _flowRuntime2.default.return(_flowRuntime2.default.ref(NodePath));

    _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType2), path);

    var subject = path;
    while (subject.type === 'QualifiedTypeIdentifier') {
      subject = subject.get('qualification');
    }
    return _flowRuntime2.default.warn(_returnType5, subject);
  }

  // First pass, collect value and type definitions

  _flowRuntime2.default.annotate(resolvePossiblyQualified, _flowRuntime2.default.function(_flowRuntime2.default.param('path', _flowRuntime2.default.ref(NodePath)), _flowRuntime2.default.return(_flowRuntime2.default.ref(NodePath))));

  (0, _babelTraverse2.default)(file, {
    ImportDeclaration(path) {
      var _pathType3 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType3), path);

      var _iteratorNormalCompletion2 = true;
      var _didIteratorError2 = false;
      var _iteratorError2 = undefined;

      try {
        for (var _iterator2 = path.get('specifiers')[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
          var specifier = _step2.value;

          var source = path.get('source').node.value;
          var local = specifier.get('local');
          var name = local.node.name;

          var vertex = registerDefinition(local);
          var imported = specifier.isImportDefaultSpecifier() ? name : specifier.get('imported').node.name;
          vertex.addDependency(graph.ref(source, imported));
        }
      } catch (err) {
        _didIteratorError2 = true;
        _iteratorError2 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion2 && _iterator2.return) {
            _iterator2.return();
          }
        } finally {
          if (_didIteratorError2) {
            throw _iteratorError2;
          }
        }
      }
    },
    ExportNamedDeclaration(path) {
      var _pathType4 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType4), path);

      if (path.node.declare) {
        // this is not a real export, it's a DeclareExportDeclaration
        // which babylon doesn't support
        registerDefinition(path.get('declaration'));
      }
    },
    TypeParameter(path) {
      var _pathType5 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType5), path);

      registerTypeParameter(path, path.parentPath.parentPath);
    },

    'DeclareTypeAlias|DeclareFunction|DeclareClass'(path) {
      var _pathType6 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType6), path);

      registerDefinition(path);
    },

    DeclareVariable(path) {
      var _pathType7 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType7), path);

      if (moduleStack.length > 1 && path.node.id.name === 'exports') {
        // Legacy feature of flow - module.exports simulated by just declaring
        // a var called exports in a module.
        // If we have no `DeclareExportDeclaration`s and no `DeclareModuleExports`,
        // convert this to a `DeclareModuleExports`
        var block = path.parentPath;
        if (block && block.isBlockStatement()) {
          var candidate = true;
          var _iteratorNormalCompletion3 = true;
          var _didIteratorError3 = false;
          var _iteratorError3 = undefined;

          try {
            for (var _iterator3 = block.get('body')[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
              var sibling = _step3.value;

              if (sibling.node === path.node) {
                continue;
              } else if (sibling.isExportNamedDeclaration() || sibling.isDeclareModuleExports()) {
                candidate = false;
                break;
              }
            }
          } catch (err) {
            _didIteratorError3 = true;
            _iteratorError3 = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion3 && _iterator3.return) {
                _iterator3.return();
              }
            } finally {
              if (_didIteratorError3) {
                throw _iteratorError3;
              }
            }
          }

          if (candidate) {
            var replacement = t.declareModuleExports();
            replacement.typeAnnotation = path.node.id.typeAnnotation;
            path.replaceWith(replacement);
            return;
          }
        }
      }
      registerDefinition(path);
    },

    DeclareModule: {
      enter(path) {
        var _pathType8 = _flowRuntime2.default.ref(NodePath);

        _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType8), path);

        var id = path.get('id');
        var name = id.isStringLiteral() ? id.node.value : id.node.name;

        var child = currentModule().registerModule(path);
        moduleStack.push(child);
      },
      exit(path) {
        var _pathType9 = _flowRuntime2.default.ref(NodePath);

        _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType9), path);

        moduleStack.pop();
      }
    },

    DeclareModuleExports(path) {
      var _pathType10 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType10), path);

      registerDefinition(path);
    },

    TypeAlias(path) {
      var _pathType11 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType11), path);

      registerDefinition(path);
      var _iteratorNormalCompletion4 = true;
      var _didIteratorError4 = false;
      var _iteratorError4 = undefined;

      try {
        for (var _iterator4 = (0, _babelPluginFlowRuntime.getTypeParameters)(path)[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
          var typeParameter = _step4.value;

          registerTypeParameter(typeParameter, path);
        }
      } catch (err) {
        _didIteratorError4 = true;
        _iteratorError4 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion4 && _iterator4.return) {
            _iterator4.return();
          }
        } finally {
          if (_didIteratorError4) {
            throw _iteratorError4;
          }
        }
      }
    },
    InterfaceDeclaration(path) {
      var _pathType12 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType12), path);

      registerDefinition(path);
      var _iteratorNormalCompletion5 = true;
      var _didIteratorError5 = false;
      var _iteratorError5 = undefined;

      try {
        for (var _iterator5 = (0, _babelPluginFlowRuntime.getTypeParameters)(path)[Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
          var typeParameter = _step5.value;

          registerTypeParameter(typeParameter, path);
        }
      } catch (err) {
        _didIteratorError5 = true;
        _iteratorError5 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion5 && _iterator5.return) {
            _iterator5.return();
          }
        } finally {
          if (_didIteratorError5) {
            throw _iteratorError5;
          }
        }
      }
    },
    VariableDeclarator(path) {
      var _pathType13 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType13), path);

      var _iteratorNormalCompletion6 = true;
      var _didIteratorError6 = false;
      var _iteratorError6 = undefined;

      try {
        for (var _iterator6 = (0, _babelPluginFlowRuntime.findIdentifiers)(path.get('id'))[Symbol.iterator](), _step6; !(_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done); _iteratorNormalCompletion6 = true) {
          var id = _step6.value;

          registerDefinition(id);
        }
      } catch (err) {
        _didIteratorError6 = true;
        _iteratorError6 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion6 && _iterator6.return) {
            _iterator6.return();
          }
        } finally {
          if (_didIteratorError6) {
            throw _iteratorError6;
          }
        }
      }
    },
    Function(path) {
      var _pathType14 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType14), path);

      var _iteratorNormalCompletion7 = true;
      var _didIteratorError7 = false;
      var _iteratorError7 = undefined;

      try {
        for (var _iterator7 = (0, _babelPluginFlowRuntime.getTypeParameters)(path)[Symbol.iterator](), _step7; !(_iteratorNormalCompletion7 = (_step7 = _iterator7.next()).done); _iteratorNormalCompletion7 = true) {
          var typeParameter = _step7.value;

          registerTypeParameter(typeParameter, path);
        }
      } catch (err) {
        _didIteratorError7 = true;
        _iteratorError7 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion7 && _iterator7.return) {
            _iterator7.return();
          }
        } finally {
          if (_didIteratorError7) {
            throw _iteratorError7;
          }
        }
      }

      if (path.isFunctionDeclaration() && path.has('id')) {
        registerDefinition(path, path.parentPath.scope);
      }

      var _iteratorNormalCompletion8 = true;
      var _didIteratorError8 = false;
      var _iteratorError8 = undefined;

      try {
        for (var _iterator8 = (0, _babelPluginFlowRuntime.findIdentifiers)(path.get('params'))[Symbol.iterator](), _step8; !(_iteratorNormalCompletion8 = (_step8 = _iterator8.next()).done); _iteratorNormalCompletion8 = true) {
          var id = _step8.value;

          registerDefinition(id);
        }
      } catch (err) {
        _didIteratorError8 = true;
        _iteratorError8 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion8 && _iterator8.return) {
            _iterator8.return();
          }
        } finally {
          if (_didIteratorError8) {
            throw _iteratorError8;
          }
        }
      }
    },
    Class(path) {
      var _pathType15 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType15), path);

      var _iteratorNormalCompletion9 = true;
      var _didIteratorError9 = false;
      var _iteratorError9 = undefined;

      try {
        for (var _iterator9 = (0, _babelPluginFlowRuntime.getTypeParameters)(path)[Symbol.iterator](), _step9; !(_iteratorNormalCompletion9 = (_step9 = _iterator9.next()).done); _iteratorNormalCompletion9 = true) {
          var typeParameter = _step9.value;

          registerTypeParameter(typeParameter, path);
        }
      } catch (err) {
        _didIteratorError9 = true;
        _iteratorError9 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion9 && _iterator9.return) {
            _iterator9.return();
          }
        } finally {
          if (_didIteratorError9) {
            throw _iteratorError9;
          }
        }
      }

      if (path.isClassDeclaration() && path.has('id')) {
        registerDefinition(path);
      }
    }
  });

  // Now walk through the graph, connecting edges
  (0, _babelTraverse2.default)(file, {
    InterfaceExtends(path) {
      var _pathType16 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType16), path);

      var id = resolvePossiblyQualified(path.get('id'));
      var name = id.node.name;

      var definition = getDefinition(id);
      var subject = path.parentPath;
      var vertex = currentModule().getEntityForPath(subject);
      if (definition) {
        vertex.addDependency(definition);
      } else {
        vertex.addDependency(graph.ref(name));
      }
    },
    GenericTypeAnnotation(path) {
      var _pathType17 = _flowRuntime2.default.ref(NodePath);

      _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType17), path);

      var id = resolvePossiblyQualified(path.get('id'));
      var name = id.node.name;

      var definition = getDefinition(id);
      var subject = findContainingPath(path);
      if (!subject) {
        console.warn('could not find a corresponding subject for node: ', path.node);
        return;
      }

      var vertex = currentModule().getEntityForPath(subject);
      if (definition) {
        vertex.addDependency(definition);
      } else {
        vertex.addDependency(graph.ref(name));
      }
    }
  });

  return graph;
}

_flowRuntime2.default.annotate(importAST, _flowRuntime2.default.function(_flowRuntime2.default.param('graph', _flowRuntime2.default.ref(FlowModule)), _flowRuntime2.default.param('file', Node)));

function findContainingPath(path) {
  var _pathType18 = _flowRuntime2.default.ref(NodePath);

  var _returnType6 = _flowRuntime2.default.return(_flowRuntime2.default.nullable(_flowRuntime2.default.ref(NodePath)));

  _flowRuntime2.default.warn(_flowRuntime2.default.param('path', _pathType18), path);

  var child = path;
  var parent = path.parentPath;
  while (parent) {
    var isFlowDeclaration = parent.type === 'DeclareVariable' || parent.type === 'DeclareClass' || parent.type === 'DeclareModule' || parent.type === 'DeclareModuleExports' || parent.type === 'DeclareFunction' || parent.type === 'DeclareTypeAlias' || parent.type === 'InterfaceDeclaration' || parent.type === 'TypeAlias' || parent.type === 'TypeCastExpression';
    if (isFlowDeclaration) {
      return _flowRuntime2.default.warn(_returnType6, parent);
    } else if (parent.isImportDefaultSpecifier() || parent.isImportNamespaceSpecifier() || path.isImportSpecifier()) {
      return _flowRuntime2.default.warn(_returnType6, child);
    } else if (parent.isStatement()) {
      return _flowRuntime2.default.warn(_returnType6, parent);
    } else {
      child = parent;
      parent = parent.parentPath;
    }
  }
}

_flowRuntime2.default.annotate(findContainingPath, _flowRuntime2.default.function(_flowRuntime2.default.param('path', _flowRuntime2.default.ref(NodePath)), _flowRuntime2.default.return(_flowRuntime2.default.nullable(_flowRuntime2.default.ref(NodePath)))));