'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _flowParser = require('flow-parser');

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _babelTypes = require('babel-types');

var t = _interopRequireWildcard(_babelTypes);

var _fastTraverse = require('./fastTraverse');

var _fastTraverse2 = _interopRequireDefault(_fastTraverse);

var _flowRuntime = require('flow-runtime');

var _flowRuntime2 = _interopRequireDefault(_flowRuntime);

var _util = require('./util');

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

var FileNode = _flowRuntime2.default.type('FileNode', _flowRuntime2.default.object(_flowRuntime2.default.property('type', _flowRuntime2.default.string()), _flowRuntime2.default.property('filename', _flowRuntime2.default.string())));

exports.default = _flowRuntime2.default.annotate(function () {
  var _ref = _asyncToGenerator(function* (filename) {
    var _filenameType = _flowRuntime2.default.string();

    var _returnType = _flowRuntime2.default.return(_flowRuntime2.default.union(FileNode, _flowRuntime2.default.ref('Promise', FileNode)));

    _flowRuntime2.default.warn(_flowRuntime2.default.param('filename', _filenameType), filename);

    var input = yield _util.fs.readFileAsync(filename, 'utf8');
    var parsed = (0, _flowParser.parse)(input, {
      esproposal_decorators: true,
      esproposal_class_instance_fields: true,
      esproposal_class_static_fields: true,
      esproposal_export_star_as: true
    });
    return _flowRuntime2.default.warn(_returnType, flowToBabylon({
      type: 'File',
      program: parsed,
      filename: _path2.default.resolve(filename)
    }));
  });

  function parseFileWithFlow(_x) {
    return _ref.apply(this, arguments);
  }

  return parseFileWithFlow;
}(), _flowRuntime2.default.function(_flowRuntime2.default.param('filename', _flowRuntime2.default.string()), _flowRuntime2.default.return(_flowRuntime2.default.ref('Promise', FileNode))));


function flowToBabylon(file) {
  var _fileType = FileNode;

  var _returnType2 = _flowRuntime2.default.return(FileNode);

  _flowRuntime2.default.warn(_flowRuntime2.default.param('file', _fileType), file);

  (0, _fastTraverse2.default)(file, function (node, key, listKey, parent) {
    if (node.type === 'ExistsTypeAnnotation') {
      node.type = 'ExistentialTypeParam';
    } else if (node.type === 'NumberLiteralTypeAnnotation') {
      node.type = 'NumericLiteralTypeAnnotation';
    } else if (node.type === 'DeclareExportDeclaration') {
      node.type = 'ExportNamedDeclaration';
      if (node.default) {
        var id = t.identifier('default');
        id.typeAnnotation = t.typeAnnotation(node.declaration);
        node.declaration = t.declareVariable(id);
      }
      node.declare = true;
    } else if (key != null && node.type === 'Identifier' && node.name === '@@iterator' && parent) {
      var replacement = t.memberExpression(t.identifier('Symbol'), t.identifier('iterator'));
      if (key === listKey) {
        parent[key] = replacement;
        parent.computed = true;
      } else if (listKey != null) {
        parent[listKey][key] = replacement;
      }
    }
  });
  return _flowRuntime2.default.warn(_returnType2, file);
}

_flowRuntime2.default.annotate(flowToBabylon, _flowRuntime2.default.function(_flowRuntime2.default.param('file', FileNode), _flowRuntime2.default.return(FileNode)));